(** This module defines dynamical strings and routines to work with them.
    Dynamical string is defined as dynamical array of chars with known length.
    0X inside the dynamical string don't interrupt it.

    For compatibility, all dynamical strings are ended with hidden 0X,
    which is not indicated in all string length parameters and functions.

    All input parameters of type ARRAY OF CHAR are processed in a full length
    rather than until the first 0X.
    
    Last change: 27.05.2015, reduced version *)

<* O2EXTENSIONS + *>
<* NOOPTIMIZE - *>

MODULE msysDynStr;

TYPE
  String * = POINTER TO ARRAY OF CHAR;
  (** Dynamical strings type *)

PROCEDURE Length * (s-:ARRAY OF CHAR): LONGINT;
  (** Returns number of symbols in s *)
  VAR
    len:LONGINT;
  BEGIN
    len:=LEN(s);
    ASSERT(len>0);

    IF s[len-1]=0X THEN
      RETURN len-1
    ELSE
      RETURN len
    END;
  END Length;

PROCEDURE New * (length:LONGINT): String;
  (** Returns a new created string with specified Length *)
  VAR
    result:String;
  BEGIN
    ASSERT(length>=0);

    NEW(result,length+1);
    result^[length]:=0X;
    RETURN result
  END New;

PROCEDURE Str * (source-:ARRAY OF CHAR): String;
  (** Returns dynamical string containing source *)
  VAR
    result:String;
    len,i:LONGINT;
  BEGIN
    len:=Length(source);
    ASSERT(len>=0);

    result:=New(len);
    FOR i:=0 TO len-1 DO
      result[i]:=source[i]
    END;

    RETURN result
  END Str;

PROCEDURE Empty * (): String;
  (** Returns the empty dynamical string *)
  BEGIN
    RETURN New(0);
  END Empty;

PROCEDURE ZeroStr * (): String;
  (** Returns dynamical string containing 0X *)
  VAR
    result:String;
  BEGIN
    result:=New(1);
    result^[0]:=0X;
    RETURN result;
  END ZeroStr;

PROCEDURE Fill * (c:CHAR; count:LONGINT): String;
  (** Returns dynamical string containing c count times *)
  VAR
    result:String;
    i:LONGINT;
  BEGIN
    ASSERT(count>=0);

    result:=New(count);
    FOR i:=0 TO count-1 DO
      result^[i]:=c
    END;
    RETURN result
  END Fill;

PROCEDURE Concat5 * (s1-,s2-,s3-,s4-,s5-:ARRAY OF CHAR): String;
  (** Returns s1+s2+s3+s4+s5 *)
  VAR
    len1,len2,len3,len4,len5,i,ind:LONGINT;
    result:String;
  BEGIN
    len1:=Length(s1);
    len2:=Length(s2);
    len3:=Length(s3);
    len4:=Length(s4);
    len5:=Length(s5);

    result:=New(len1+len2+len3+len4+len5);

    ind:=0;
    FOR i:=0 TO len1-1 DO
      result[ind]:=s1[i];
      INC(ind)
    END;
    FOR i:=0 TO len2-1 DO
      result[ind]:=s2[i];
      INC(ind)
    END;
    FOR i:=0 TO len3-1 DO
      result[ind]:=s3[i];
      INC(ind)
    END;
    FOR i:=0 TO len4-1 DO
      result[ind]:=s4[i];
      INC(ind)
    END;
    FOR i:=0 TO len5-1 DO
      result[ind]:=s5[i];
      INC(ind)
    END;

    RETURN result
  END Concat5;

PROCEDURE Concat4 * (s1-,s2-,s3-,s4-:ARRAY OF CHAR): String;
  (** Returns s1+s2+s3+s4 *)
  BEGIN
    RETURN Concat5(s1,s2,s3,s4,'');
  END Concat4;

PROCEDURE Concat3 * (s1-,s2-,s3-:ARRAY OF CHAR): String;
  (** Returns s1+s2+s3 *)
  BEGIN
    RETURN Concat5(s1,s2,s3,'','')
  END Concat3;

PROCEDURE Concat2 * (s1-,s2-:ARRAY OF CHAR): String;
  (** Returns s1+s2 *)
  BEGIN
    RETURN Concat5(s1,s2,'','','')
  END Concat2;

PROCEDURE Copy * (s-:ARRAY OF CHAR; Index,Count:LONGINT): String;
  (** Returns a substring containing Count characters starting at s[Index].
      The substring is a unique copy (that is, it does not share memory with s).
      If Index is larger than the length of s, Copy returns an empty string.
      If Count specifies more characters or array elements than are available,
      only the characters or elements from s[Index] to the end of s are returned. *)
  VAR
    len,i,LenS:LONGINT;
    result:String;
  BEGIN
    ASSERT(Index>=0);
    ASSERT(Count>=0);

    LenS:=Length(s);

    IF Index>LenS-1 THEN
      RETURN Empty();
    END;

    IF Count>LenS-Index THEN
      len:=LenS-Index
    ELSE
      len:=Count
    END;

    result:=New(len);
    FOR i:=0 TO len-1 DO
      result[i]:=s[Index+i]
    END;

    RETURN result
  END Copy;

PROCEDURE Delete * (s-:ARRAY OF CHAR; Index,Count:LONGINT): String;
  (** Removes a substring of Count characters from string s starting with s[Index]
      and returns the result. If index is larger than the length of the string or
      less than 0, no characters are deleted. If count specifies more characters
      than remain starting at the index, Delete removes the rest of the string.
      If count is less than or equal to 0, no characters are deleted. *)
  VAR
    result:String;
    i,LenS,len:LONGINT;
  BEGIN
    LenS:=Length(s);

    IF (Index>LenS-1) OR (Count<=0) THEN
      RETURN Copy(s,0,LenS);
    END;

    IF Index+Count>LenS THEN
      len:=Index
    ELSE
      len:=LenS-Count
    END;

    result:=New(len);

    FOR i:=0 TO Index-1 DO
      result[i]:=s[i];
    END;

    FOR i:=Index+Count TO LenS-1 DO
      result[i-Count]:=s[i]
    END;

    RETURN result
  END Delete;

PROCEDURE Insert * (SubString-,s-:ARRAY OF CHAR; Index:LONGINT): String;
  (** Inserts SubString into s at the position s[index].
      If Index is less than 0, it is mapped to 0. If it is past the end
      of the string, it is set to the length of the string, turning
      the operation into an append. If the SubString parameter is
      an empty string, Insert does nothing. *)
  VAR
    result:String;
    i,j,k,len,LenSubString,LenS: LONGINT;
  BEGIN
    LenSubString:=Length(SubString);
    LenS:=Length(s);
    len:=LenSubString+LenS;

    result:=New(len);

    i:=0;
    WHILE (i<LenS) & (i<Index) DO
      result[i]:=s[i];
      INC(i)
    END;

    j:=0;
    k:=i;
    WHILE j<LenSubString DO
      result[i]:=SubString[j];
      INC(i);
      INC(j)
    END;

    WHILE i<len DO
      result[i]:=s[k];
      INC(i);
      INC(k);
    END;

    RETURN result
  END Insert;

PROCEDURE Pos * (SubString-,s-:ARRAY OF CHAR): LONGINT;
  (** Searches for SubString within s and returns an integer value
      that is the index of the first character of Substr within s.
      Pos is case-sensitive. If SubString is not found, Pos returns -1. *)
  VAR
    result,LenS,LenSubString:LONGINT;
  PROCEDURE IsEqual (): BOOLEAN;
    VAR
      b:BOOLEAN;
      i:LONGINT;
    BEGIN
      b:=TRUE;
      i:=0;
      WHILE b & (i<LenSubString) DO
        b:=(SubString[i]=s[i+result]);
        INC(i);
      END;
      RETURN b
    END IsEqual;
  BEGIN
    LenSubString:=Length(SubString);
    IF LenSubString=0 THEN
      RETURN -1
    END;
    LenS:=Length(s);

    result:=0;
    WHILE (result<=LenS-LenSubString) & (~IsEqual()) DO
      INC(result)
    END;

    IF result>LenS-LenSubString THEN
      result:=-1
    END;

    RETURN result
  END Pos;

PROCEDURE DoCompare (s1-,s2-:ARRAY OF CHAR; p:PROCEDURE (c:CHAR): CHAR): LONGINT;
  VAR
    Len1,Len2,MinLen:LONGINT;
    i,result:LONGINT;
  BEGIN
    ASSERT(p#NIL);

    Len1:=Length(s1);
    Len2:=Length(s2);
    IF (Len1=0) OR (Len2=0) THEN
      RETURN Len1-Len2
    END;

    IF Len1<Len2 THEN
      MinLen:=Len1
    ELSE
      MinLen:=Len2
    END;

    i:=0;
    REPEAT
      result:=ORD(p(s1[i]))-ORD(p(s2[i]));
      INC(i)
    UNTIL (result#0) OR (i=MinLen);

    IF result=0 THEN
      RETURN Len1-Len2
    ELSE
      RETURN result
    END;
  END DoCompare;

PROCEDURE CharUnity (c:CHAR): CHAR;
  BEGIN
    RETURN c
  END CharUnity;

PROCEDURE Compare * (s1-,s2-:ARRAY OF CHAR): LONGINT;
  (** Compares s1 to s2. The return value is less than 0 if s1 is less than s2,
      0 if s1 equals s2, or greater than 0 if s1 is greater than s2 *)
  BEGIN
    RETURN DoCompare(s1,s2,CharUnity)
  END Compare;

PROCEDURE Capitalize (c:CHAR): CHAR;
  BEGIN
    RETURN CAP(c)
  END Capitalize;

PROCEDURE CompareCapitalized * (s1-,s2-:ARRAY OF CHAR): LONGINT;
  (** Compares CAP(s1) to CAP(s2). The return value is less than 0 if CAP(s1) is
      less than CAP(s2), 0 if CAP(s1) equals CAP(s2),
      or greater than 0 if CAP(s1) is greater than CAP(s2) *)
  BEGIN
    RETURN DoCompare(s1,s2,Capitalize);
  END CompareCapitalized;

PROCEDURE IsSpace * (c:CHAR): BOOLEAN;
  (** Returns if the char is space, tab, end of line, or some other symbol
      which can be skipped until a number in the text stream *)
  BEGIN
    RETURN (ORD(c)<=32)
  END IsSpace;

PROCEDURE LenTill0 * (source-:ARRAY OF CHAR): LONGINT;
  (** Returns length of the source till the first 0X.
      The 0X is not counted. *)
  VAR
    len:LONGINT;
  BEGIN
    len:=0;
    WHILE (len<LEN(source)) & (source[len]#0X) DO
      INC(len)
    END;

    RETURN len;
  END LenTill0;

PROCEDURE Till0 * (source-:ARRAY OF CHAR): String;
  (** Returns dynamical string containing the source till the first 0X. *)
  VAR
    i,len:LONGINT;
    result:String;
  BEGIN
    len:=LenTill0(source);

    result:=New(len);
    FOR i:=0 TO len-1 DO
      result[i]:=source[i]
    END;

    RETURN result
  END Till0;

END msysDynStr.